package com.ippon.borestop.common.domain.error;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

/**
 * Parent exception used in Borestop application. Those exceptions will be resolved as human readable errors.
 *
 * <p>
 * You can use this implementation directly:
 * </p>
 *
 * <p>
 * <code>
 *     <pre>
 *       BorestopException.builder(StandardMessages.USER_MANDATORY)
 *          .argument("key", "value")
 *          .argument("other", "test")
 *          .status(ErrorsHttpStatus.BAD_REQUEST)
 *          .message("Error message")
 *          .cause(new RuntimeException())
 *          .build();
 *     </pre>
 *   </code>
 * </p>
 *
 * <p>
 * Or make extension exceptions:
 * </p>
 *
 * <p>
 * <code>
 *     <pre>
 *       public class MissingMandatoryValueException extends BorestopException {
 *
 *         public MissingMandatoryValueException(BorestopMessage borestopMessage, String fieldName) {
 *           this(builder(borestopMessage, fieldName, defaultMessage(fieldName)));
 *         }
 *
 *         protected MissingMandatoryValueException(BorestopExceptionBuilder builder) {
 *           super(builder);
 *         }
 *
 *         private static BorestopExceptionBuilder builder(BorestopMessage borestopMessage, String fieldName, String message) {
 *           return BorestopException.builder(borestopMessage)
 *               .status(ErrorsHttpStatus.INTERNAL_SERVER_ERROR)
 *               .argument("field", fieldName)
 *               .message(message);
 *         }
 *
 *         private static String defaultMessage(String fieldName) {
 *           return "The field \"" + fieldName + "\" is mandatory and wasn't set";
 *         }
 *       }
 *     </pre>
 *   </code>
 * </p>
 */
public class BorestopException extends RuntimeException {

  private final Map<String, String> arguments;
  private final ErrorStatus status;
  private final BorestopMessage borestopMessage;

  protected BorestopException(BorestopExceptionBuilder builder) {
    super(getMessage(builder), getCause(builder));
    arguments = getArguments(builder);
    status = getStatus(builder);
    borestopMessage = getBorestopMessage(builder);
  }

  private static String getMessage(BorestopExceptionBuilder builder) {
    if (builder == null) {
      return null;
    }

    return builder.message;
  }

  private static Throwable getCause(BorestopExceptionBuilder builder) {
    if (builder == null) {
      return null;
    }

    return builder.cause;
  }

  private static Map<String, String> getArguments(BorestopExceptionBuilder builder) {
    if (builder == null) {
      return null;
    }

    return Collections.unmodifiableMap(builder.arguments);
  }

  private static ErrorStatus getStatus(BorestopExceptionBuilder builder) {
    if (builder == null) {
      return null;
    }

    return builder.status;
  }

  private static BorestopMessage getBorestopMessage(BorestopExceptionBuilder builder) {
    if (builder == null) {
      return null;
    }

    return builder.borestopMessage;
  }

  public static BorestopExceptionBuilder builder(BorestopMessage message) {
    return new BorestopExceptionBuilder(message);
  }

  public Map<String, String> getArguments() {
    return arguments;
  }

  public ErrorStatus getStatus() {
    return status;
  }

  public BorestopMessage getBorestopMessage() {
    return borestopMessage;
  }

  public static class BorestopExceptionBuilder {

    private final Map<String, String> arguments = new HashMap<>();
    private String message;
    private ErrorStatus status;
    private BorestopMessage borestopMessage;
    private Throwable cause;

    public BorestopExceptionBuilder(BorestopMessage borestopMessage) {
      this.borestopMessage = borestopMessage;
    }

    public BorestopExceptionBuilder argument(String key, Object value) {
      arguments.put(key, getStringValue(value));

      return this;
    }

    private String getStringValue(Object value) {
      if (value == null) {
        return "null";
      }

      return value.toString();
    }

    public BorestopExceptionBuilder status(ErrorStatus status) {
      this.status = status;

      return this;
    }

    public BorestopExceptionBuilder message(String message) {
      this.message = message;

      return this;
    }

    public BorestopExceptionBuilder cause(Throwable cause) {
      this.cause = cause;

      return this;
    }

    public BorestopException build() {
      return new BorestopException(this);
    }
  }
}
