package com.ippon.borestop.common.infrastructure.primary;

import static org.assertj.core.api.Assertions.assertThat;

import com.ippon.borestop.common.domain.error.BorestopMessage;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.stream.Collectors;
import org.junit.jupiter.api.Test;
import org.reflections.Reflections;
import org.reflections.scanners.SubTypesScanner;
import org.reflections.util.ClasspathHelper;
import org.reflections.util.ConfigurationBuilder;
import org.reflections.util.FilterBuilder;

public class ErrorMessagesUnitTest {

  private static final Set<Class<? extends BorestopMessage>> errors = new Reflections(
    new ConfigurationBuilder()
      .setUrls(ClasspathHelper.forPackage("com.ippon"))
      .setScanners(new SubTypesScanner())
      .filterInputsBy(new FilterBuilder().includePackage("com.ippon"))
  )
    .getSubTypesOf(BorestopMessage.class);

  @Test
  public void shouldHaveOnlyEnumImplementations() {
    errors.forEach(
      error ->
        assertThat(error.isEnum() || error.isInterface())
          .as("Implementations of " + BorestopMessage.class.getName() + " must be enums and " + error.getName() + " wasn't")
          .isTrue()
    );
  }

  @Test
  public void shouldHaveMessagesForAllKeys() {
    Collection<Properties> messages = loadMessages();

    errors
      .stream()
      .filter(error -> error.isEnum())
      .forEach(error -> Arrays.stream(error.getEnumConstants()).forEach(value -> messages.forEach(assertMessageExist(value))));
  }

  private List<Properties> loadMessages() {
    try {
      return Files.list(Paths.get("src/main/resources/i18n")).map(toProperties()).collect(Collectors.toList());
    } catch (IOException e) {
      throw new AssertionError();
    }
  }

  private Function<Path, Properties> toProperties() {
    return file -> {
      Properties properties = new Properties();
      try (InputStream input = Files.newInputStream(file)) {
        properties.load(input);
      } catch (IOException e) {
        throw new AssertionError();
      }

      return properties;
    };
  }

  private Consumer<Properties> assertMessageExist(BorestopMessage value) {
    return currentMessages -> {
      assertThat(currentMessages.getProperty("borestop.error." + value.getMessageKey()))
        .as("Can't find message for " + value.getMessageKey() + " in all files, check your configuration")
        .isNotBlank();
    };
  }
}
