package fr.ippon.life;

import java.util.Collection;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

public class Grid {

  private final Set<Cell> aliveCells;
  private final Set<Cell> deadCells;
  private final Box border;

  Grid(Set<Cell> aliveCells) {
    this.aliveCells = aliveCells;
    this.deadCells = buildDeadCells(aliveCells);
    border = new Box(aliveCells);
  }

  private Set<Cell> buildDeadCells(Set<Cell> aliveCells) {
    return aliveCells
      .stream()
      .map(Box::new)
      .flatMap(Box::cells)
      .filter(cell -> !aliveCells.contains(cell))
      .collect(Collectors.toUnmodifiableSet());
  }

  public boolean isAlive(Cell cell) {
    return aliveCells.contains(cell);
  }

  public Collection<Cell> deadCells() {
    return deadCells;
  }

  public int leftColumn() {
    return border.leftColumn();
  }

  public int rightColumn() {
    return border.rightColumn();
  }

  public int topRow() {
    return border.topRow();
  }

  public int bottomRow() {
    return border.bottomRow();
  }

  private static class Box {

    private final Point topLeft;
    private final Point bottomRight;

    public Box(Cell cell) {
      this(Set.of(cell));
    }

    public Box(Set<Cell> cells) {
      topLeft = Point.topLeft(cells);
      bottomRight = Point.bottomRight(cells);
    }

    private Stream<Cell> cells() {
      return IntStream.range(leftColumn(), rightColumn()).mapToObj(Integer::valueOf).flatMap(column -> cellsColumn(column));
    }

    private Stream<Cell> cellsColumn(Integer column) {
      return IntStream.range(topRow(), bottomRow()).mapToObj(Integer::valueOf).map(row -> new Cell(row, column));
    }

    public int topRow() {
      return topLeft.getRow();
    }

    public int bottomRow() {
      return bottomRight.getRow() + 1;
    }

    public int leftColumn() {
      return topLeft.getColumn();
    }

    public int rightColumn() {
      return bottomRight.getColumn() + 1;
    }
  }

  private static class Point {

    private final int row;
    private final int column;

    public Point(int row, int column) {
      this.row = row;
      this.column = column;
    }

    private static Point topLeft(Set<Cell> cells) {
      int row = cells.stream().mapToInt(Cell::row).min().orElse(0);

      int column = cells.stream().mapToInt(Cell::column).min().orElse(0);

      return new Point(row - 1, column - 1);
    }

    private static Point bottomRight(Set<Cell> cells) {
      int row = cells.stream().mapToInt(Cell::row).max().orElse(0);

      int column = cells.stream().mapToInt(Cell::column).max().orElse(0);

      return new Point(row + 1, column + 1);
    }

    public int getRow() {
      return row;
    }

    public int getColumn() {
      return column;
    }
  }
}
