package fr.craft.kata;

import static org.assertj.core.api.Assertions.*;

import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.HashSet;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import org.junit.jupiter.api.Test;

public class ThreadingTest {

  @Test
  void shouldStopThread() throws InterruptedException {
    StopableThread process = new StopableThread();
    Thread thread = new Thread(process);
    thread.start();

    Thread.sleep(100);
    process.stop();

    Thread.sleep(200);

    assertThat(thread.isAlive()).isFalse();
  }

  @Test
  void shouldPlayWithThreads() throws InterruptedException {
    ExecutorService executor = Executors.newFixedThreadPool(5);

    AtomicInteger counter = new AtomicInteger();
    Set<Future<Integer>> results = new HashSet<>();
    for (int i = 0; i < 100; i++) {
      results.add(executor.submit(() -> counter.incrementAndGet()));
    }

    executor.shutdown();
    executor.awaitTermination(1, TimeUnit.SECONDS);

    assertThat(results).extracting(result -> result.get()).hasSize(100);
  }

  @Test
  void shouldHaveThreadSafeCollection() throws InterruptedException {
    ExecutorService executor = Executors.newFixedThreadPool(5);

    Collection<UUID> ids = Collections.newSetFromMap(new ConcurrentHashMap<>());
    for (int i = 0; i < 100; i++) {
      executor.submit(() -> ids.add(UUID.randomUUID()));
    }

    executor.shutdown();
    executor.awaitTermination(1, TimeUnit.SECONDS);

    assertThat(ids).hasSize(100);
  }

  @Test
  void shouldHaveOldDates() throws InterruptedException {
    ExecutorService executor = Executors.newFixedThreadPool(5);

    Collection<Integer> days = Collections.newSetFromMap(new ConcurrentHashMap<>());
    for (int day = 0; day < 31; day++) {
      GregorianCalendar date = new GregorianCalendar(2021, 00, day);

      executor.submit(
        () -> {
          SimpleDateFormat format = new SimpleDateFormat("dd");

          return days.add(Integer.parseInt(format.format(date.getTime())));
        }
      );
    }

    executor.shutdown();
    executor.awaitTermination(1, TimeUnit.SECONDS);

    assertThat(days).hasSize(31);
  }

  private static class StopableThread implements Runnable {

    private AtomicBoolean running = new AtomicBoolean(true);

    @Override
    public void run() {
      while (running.get()) {}
    }

    public void stop() {
      running.set(false);
    }
  }
}
