type Color = 'purple' | 'pink' | 'red' | 'blue' | 'yellow' | 'green' | 'white' | 'brown';

type Colors = [Color, Color, Color, Color];

type State = 'playing' | 'lost' | 'win';

interface Report {
  wellPlaced: number;
  misplaced: number;
}

const difference = (colors: Color[], substracts: Color[]) => {
  const remaining = [...substracts];

  return colors.filter((current) => {
    if (remaining.includes(current)) {
      const toRemove = remaining.indexOf(current);
      remaining.splice(toRemove, 1);
      return false;
    }
    return true;
  });
};

const intersect = (secret: Color[], guess: Color[]) => {
  const remaining = [...guess];

  return secret.filter((current) => {
    if (remaining.includes(current)) {
      const toRemove = remaining.indexOf(current);
      remaining.splice(toRemove, 1);
      return true;
    }
    return false;
  });
};

class Mastermind {
  private currentState: State = 'playing';
  private remaining = 12;
  constructor(private secret: Colors) {}

  evaluate(guess: Colors): Report {
    this.remaining--;
    const wellPlaced = this.listWellPlaced(guess);
    const rest = difference(this.secret, wellPlaced);
    const guessRest = difference(guess, wellPlaced);
    const misplaced = intersect(rest, guessRest).length;
    const wellPlacedCount = this.countWellPlaced(guess);
    this.currentState = this.evaluateState(wellPlacedCount);
    return {
      wellPlaced: wellPlacedCount,
      misplaced,
    };
  }

  private evaluateState(wellPlacedCount: number): State {
    if (this.remaining <= 0) {
      return 'lost';
    }
    if (wellPlacedCount === 4) {
      return 'win';
    }
    return 'playing';
  }

  private listWellPlaced(guess: Color[]): Color[] {
    return this.secret.filter((current, index) => current === guess[index]);
  }

  private countWellPlaced(guess: Color[]): number {
    return this.listWellPlaced(guess).length;
  }

  get state(): State {
    return this.currentState;
  }
}

const defaultMastermind = () => new Mastermind(['purple', 'pink', 'red', 'blue']);

const evaluateToLose = (mastermind: Mastermind) => {
  for (let i = 0; i < 12; i++) {
    mastermind.evaluate(['red', 'red', 'red', 'red']);
  }
};

describe('Mastermind', () => {
  describe('Evaluation', () => {
    it('Should have all color well placed', () => {
      const mastermind = defaultMastermind();

      expect(mastermind.evaluate(['purple', 'pink', 'red', 'blue'])).toEqual<Report>({
        wellPlaced: 4,
        misplaced: 0,
      });
    });

    it('Should not have any well placed', () => {
      const mastermind = defaultMastermind();

      expect(mastermind.evaluate(['yellow', 'green', 'white', 'brown'])).toEqual<Report>({
        wellPlaced: 0,
        misplaced: 0,
      });
    });

    it('Should have only misplaced', () => {
      expect(defaultMastermind().evaluate(['pink', 'purple', 'blue', 'red'])).toEqual<Report>({
        wellPlaced: 0,
        misplaced: 4,
      });
    });

    it('Should have two colors misplaced and two colors well placed', () => {
      const mastermind = new Mastermind(['purple', 'pink', 'purple', 'yellow']);

      expect(mastermind.evaluate(['purple', 'pink', 'yellow', 'purple'])).toEqual<Report>({
        wellPlaced: 2,
        misplaced: 2,
      });
    });

    it('Should not misplace repeated secret', () => {
      const mastermind = new Mastermind(['red', 'red', 'red', 'red']);

      expect(mastermind.evaluate(['red', 'yellow', 'yellow', 'yellow'])).toEqual<Report>({
        wellPlaced: 1,
        misplaced: 0,
      });
    });
  });

  describe('State', () => {
    it('Should be pending by default', () => {
      expect(defaultMastermind().state).toBe<State>('playing');
    });

    it('Should lost after 12 tries', () => {
      const mastermind = defaultMastermind();

      for (let i = 0; i < 12; i++) {
        mastermind.evaluate(['red', 'red', 'red', 'red']);
      }

      expect(mastermind.state).toBe<State>('lost');
    });

    it('Should win when already well placed', () => {
      const mastermind = defaultMastermind();

      mastermind.evaluate(['purple', 'pink', 'red', 'blue']);

      expect(mastermind.state).toBe<State>('win');
    });

    it("Can't win when already lost", () => {
      const mastermind = defaultMastermind();

      evaluateToLose(mastermind);

      mastermind.evaluate(['purple', 'pink', 'red', 'blue']);

      expect(mastermind.state).toBe<State>('lost');
    });
  });

  describe('Difference', () => {
    it('Should not remove duplicate items', () => {
      expect(difference(['purple', 'purple'], ['purple', 'pink'])).toEqual(['purple']);
    });
  });
});
