package fr.ippon.kata.puissance4;

import static org.assertj.core.api.Assertions.*;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.io.StringReader;
import java.util.Arrays;
import java.util.NoSuchElementException;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;
import org.awaitility.Awaitility;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

class Puissance4Test {

  private PrintStream systemOut;
  private InputStream systemIn;

  private ByteArrayOutputStream out;
  private SystemInMock in = new SystemInMock();
  private Thread runner;

  @BeforeEach
  void changeOutput() {
    systemOut = System.out;
    systemIn = System.in;

    out = new ByteArrayOutputStream();
    System.setOut(new PrintStream(out));
    System.setIn(in);
  }

  @AfterEach
  void resetOutput() {
    System.setOut(systemOut);
    System.setIn(systemIn);
  }

  @AfterEach
  void stopGame() {
    runner.interrupt();
  }

  @Test
  void shouldDisplayMenu() {
    play();

    assertGameOrder("Puissance 4", "[3] Regle du jeu");
  }

  @Test
  void shouldDisplayRules() throws InterruptedException {
    play("3", "1");

    assertGameOrder("[3] Regle du jeu", "Le but du jeu", "la partie est déclarée nulle", "[3] Regle du jeu");
  }

  @Test
  void shouldPlayTwoPlayerGame() {
    play("2", "Colin", "Cyril", "1", "2", "1", "2", "1", "2", "1");

    assertGameOrder("Colin a gagné !");
  }

  @Test
  void shouldAskMenuAgainForOutOfBoundOption() {
    play("5", "0", "3");

    assertGameOrder("[3] Regle du jeu");
  }

  @Test
  void shouldAskColumnAgainForBadPlayerOneInput() {
    play("2", "Colin", "Cyril", "0", "9", "1");

    assertGameOrder(
      "Choix incorrect, entrez un nombre entre 1 et 8",
      "Choix incorrect, entrez un nombre entre 1 et 8",
      "| X| *| *| *| *| *| *| *|"
    );
  }

  @Test
  void shouldAskColumnAgainForBadPlayerTwoInput() {
    play("2", "Colin", "Cyril", "1", "9", "0", "2");

    assertGameOrder(
      "Choix incorrect, entrez un nombre entre 1 et 8",
      "Choix incorrect, entrez un nombre entre 1 et 8",
      "| X| O| *| *| *| *| *| *|"
    );
  }

  @Test
  void shouldPlayAgainsIA() {
    play("1", "Colin", "4", "5", "6", "7", "8");

    assertGameOrder(
      "| *| *| *| X| *| *| *| *|",
      "| *| *| *| X| O| *| *| *|",
      "| *| *| *| *| X| *| *| *|",
      "| *| *| *| O| X| *| *| *|",
      "| *| *| *| X| O| X| *| *|",
      "| *| *| *| O| *| *| *| *|",
      "L'ordinateur a gagné !"
    );
  }

  private void play(String... inputs) {
    in.provideText(Arrays.stream(inputs).collect(Collectors.joining("\n")));

    runner = startGame();
  }

  private Thread startGame() {
    Thread runner = new Thread(
      () -> {
        try {
          Puissance4.main(null);
        } catch (NoSuchElementException e) {
          // Nothing to do
        }
      }
    );
    runner.start();

    return runner;
  }

  private void assertGameOrder(String... elements) {
    Awaitility
      .await()
      .pollDelay(5, TimeUnit.MILLISECONDS)
      .atMost(1000, TimeUnit.MILLISECONDS)
      .untilAsserted(
        () -> {
          String message = new String(out.toByteArray()).intern();

          for (String element : elements) {
            assertThat(message).contains(element);
            message = message.substring(message.indexOf(element) + element.length());
          }
        }
      );
  }

  private static class SystemInMock extends InputStream {

    private StringReader currentReader;
    private IOException ioException;
    private RuntimeException runtimeException;

    void provideText(String text) {
      currentReader = new StringReader(text);
    }

    @Override
    public int read() throws IOException {
      if (currentReader == null) {
        return 0;
      }

      int character = currentReader.read();
      if (character == -1) handleEmptyReader();
      return character;
    }

    private void handleEmptyReader() throws IOException {
      if (ioException != null) throw ioException; else if (runtimeException != null) throw runtimeException;
    }

    @Override
    public int read(byte[] buffer, int offset, int len) throws IOException {
      if (buffer == null) throw new NullPointerException(); else if (
        offset < 0 || len < 0 || len > buffer.length - offset
      ) throw new IndexOutOfBoundsException(); else if (len == 0) return 0; else return readNextLine(buffer, offset, len);
    }

    private int readNextLine(byte[] buffer, int offset, int len) throws IOException {
      int c = read();
      if (c == -1) return -1;
      buffer[offset] = (byte) c;

      int i = 1;
      for (; (i < len) && !isCompleteLineWritten(buffer, i - 1); ++i) {
        byte read = (byte) read();
        if (read == -1) break; else buffer[offset + i] = read;
      }
      return i;
    }

    private boolean isCompleteLineWritten(byte[] buffer, int indexLastByteWritten) {
      byte[] separator = System.lineSeparator().getBytes();
      int indexFirstByteOfSeparator = indexLastByteWritten - separator.length + 1;
      return indexFirstByteOfSeparator >= 0 && contains(buffer, separator, indexFirstByteOfSeparator);
    }

    private boolean contains(byte[] array, byte[] pattern, int indexStart) {
      for (int i = 0; i < pattern.length; ++i) if (array[indexStart + i] != pattern[i]) return false;
      return true;
    }
  }
}
