const ARABIC_TO_ROMAN = {
  1: 'I',
  4: 'IV',
  5: 'V',
  9: 'IX',
  10: 'X',
  40: 'XL',
  50: 'L',
  90: 'XC',
  100: 'C',
  400: 'CD',
  500: 'D',
  900: 'CM',
  1000: 'M',
};

const toRoman = (arabic: number): string => {
  if (arabic === 0) {
    return '';
  }
  const [foundArabic, foundRoman] = Object.entries(ARABIC_TO_ROMAN)
    .reverse()
    .find(([number]) => arabic >= +number);
  return foundRoman.concat(toRoman(arabic - +foundArabic));
};

const toArabic = (roman: string): number => {
  if (roman.length === 0) {
    return 0;
  }
  const [foundArabic, foundRoman] = Object.entries(ARABIC_TO_ROMAN)
    .reverse()
    .find(([, romanNumber]) => roman.startsWith(romanNumber));
  return +foundArabic + toArabic(roman.substring(foundRoman.length));
};

const romanSum = (first: string, second: string) => toRoman(toArabic(first) + toArabic(second));

const romanMultiply = (first: string, second: string): string => toRoman(toArabic(first) * toArabic(second));

const romanOperation = (operation: string): string => {
  const [first, second] = operation.split(/[+*]/).map((string) => string.trim());
  if (operation.includes('*')) {
    return romanMultiply(first, second);
  }
  return romanSum(first, second);
};

describe('Roman Calculator', () => {
  describe('For Arabic', () => {
    it('Should be empty for 0', () => {
      expect(toRoman(0)).toBe('');
    });

    it('Should be I for 1', () => {
      expect(toRoman(1)).toBe('I');
    });

    it('Should be III for 3', () => {
      expect(toRoman(3)).toBe('III');
    });

    it('Should be IV for 4', () => {
      expect(toRoman(4)).toBe('IV');
    });

    it('Should be V for 5', () => {
      expect(toRoman(5)).toBe('V');
    });

    it('Should be VI for 6', () => {
      expect(toRoman(6)).toBe('VI');
    });

    it('Should be VIII for 8', () => {
      expect(toRoman(8)).toBe('VIII');
    });

    it('Should be IX for 9', () => {
      expect(toRoman(9)).toBe('IX');
    });

    it('Should be X for 10', () => {
      expect(toRoman(10)).toBe('X');
    });

    it('Should be XL for 40', () => {
      expect(toRoman(40)).toBe('XL');
    });

    it('Should be L for 50', () => {
      expect(toRoman(50)).toBe('L');
    });

    it('Should be XC for 90', () => {
      expect(toRoman(90)).toBe('XC');
    });

    it('Should be C for 100', () => {
      expect(toRoman(100)).toBe('C');
    });

    it('Should be CD for 400', () => {
      expect(toRoman(400)).toBe('CD');
    });

    it('Should be D for 500', () => {
      expect(toRoman(500)).toBe('D');
    });

    it('Should be CM for 900', () => {
      expect(toRoman(900)).toBe('CM');
    });

    it('Should be M for 1000', () => {
      expect(toRoman(1000)).toBe('M');
    });

    it('Should be MMMMDCCCLXXXVIII for 4888', () => {
      expect(toRoman(4888)).toBe('MMMMDCCCLXXXVIII');
    });

    it('Should be MCMXCVIII for 1998', () => {
      expect(toRoman(1998)).toBe('MCMXCVIII');
    });
  });
  describe('For Roman', () => {
    it('Should be 0 for empty string', () => {
      expect(toArabic('')).toBe(0);
    });

    it('Should be 1 for I', () => {
      expect(toArabic('I')).toBe(1);
    });

    it('Should be 3 for III', () => {
      expect(toArabic('III')).toBe(3);
    });

    it('Should be 4 for IV', () => {
      expect(toArabic('IV')).toBe(4);
    });

    it('Should be 5 for V', () => {
      expect(toArabic('V')).toBe(5);
    });

    it('Should be 6 for VI', () => {
      expect(toArabic('VI')).toBe(6);
    });

    it('Should be 1998 for MCMXCVIII', () => {
      expect(toArabic('MCMXCVIII')).toBe(1998);
    });
  });

  describe('Calculate', () => {
    it('Should sum III and IV', () => {
      expect(romanSum('III', 'IV')).toBe('VII');
    });

    it('Should multiply VI and VII', () => {
      expect(romanMultiply('VI', 'VII')).toBe('XLII');
    });

    it('Should sum VIII and I', () => {
      expect(romanOperation('VIII + I')).toBe('IX');
    });

    it('Should sum IX and IX', () => {
      expect(romanOperation('IX * IX')).toBe('LXXXI');
    });
  });
});
