package fr.ippon.craft;

import java.util.Objects;
import java.util.function.Function;

public abstract class Try<T> {

  public static <T> Try<T> of(Faillable<T> failable) {
    try {
      T result = failable.execute();

      return new SuccessfulTry<>(result);
    } catch (Throwable e) {
      return new FailingTry<>(e);
    }
  }

  public abstract <R> Try<R> map(Function<T, R> mapper);

  public abstract <R> Try<R> flatMap(Function<T, Try<R>> failable);

  public abstract T get();

  public abstract T orElse(T recovery);

  public abstract T orElseGet(Faillable<T> failing);

  private static class SuccessfulTry<T> extends Try<T> {

    private final T result;

    private SuccessfulTry(T result) {
      this.result = result;
    }

    @Override
    public <R> Try<R> map(Function<T, R> mapper) {
      Objects.requireNonNull(mapper);

      return new SuccessfulTry<R>(mapper.apply(result));
    }

    @Override
    public <R> Try<R> flatMap(Function<T, Try<R>> failable) {
      return failable.apply(result);
    }

    @Override
    public T get() {
      return result;
    }

    @Override
    public T orElse(T recovery) {
      return result;
    }

    @Override
    public T orElseGet(Faillable<T> failing) {
      return result;
    }
  }

  private static class FailingTry<T> extends Try<T> {

    private final Throwable e;

    private FailingTry(Throwable e) {
      this.e = e;
    }

    @Override
    public <R> Try<R> map(Function<T, R> mapper) {
      return new FailingTry<>(e);
    }

    @Override
    public <R> Try<R> flatMap(Function<T, Try<R>> failable) {
      return new FailingTry<>(e);
    }

    @Override
    public T get() {
      throw new ExecutionException(e);
    }

    @Override
    public T orElse(T recovery) {
      return recovery;
    }

    @Override
    public T orElseGet(Faillable<T> failing) {
      return Try.of(failing).get();
    }
  }
}
